################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import abc
from dataclasses import dataclass
from datetime import timedelta
from keyword import iskeyword, kwlist


@dataclass(repr=True, eq=True, order=False, frozen=True)
class SdkAddress:
    namespace: str
    name: str
    id: str
    typename: str


class TypeSerializer(metaclass=abc.ABCMeta):
    __slots__ = ()

    """
    A base class for a TypeSerializer. A TypeSerializer is responsible of serialising and deserializing a specific
    value type.
    """

    @abc.abstractmethod
    def serialize(self, value) -> bytes:
        """
        Serialize the given value to bytes.

        :param value: the value to serialize.
        :return: a byte representation of the given value.
        """
        pass

    @abc.abstractmethod
    def deserialize(self, byte: bytes):
        """
        Deserialize a value from the given byte representation.

        :param byte: the bytes that represent the value to deseralize.
        :return: the deserialized value.
        """
        pass


class Type(metaclass=abc.ABCMeta):
    __slots__ = ("typename",)

    """
    A base representation of a Stateful Function value type.
    """

    def __init__(self, typename: str):
        """
        :param typename: a TypeName represented as a string of the form <namespace>/<name> of this type.
        """
        if not typename:
            raise ValueError("typename can not be missing")
        self.typename = typename

    @abc.abstractmethod
    def serializer(self) -> TypeSerializer:
        """
        :return: a serializer for this type.
        """
        pass


class ValueSpec(object):
    __slots__ = ("name", "type", "duration", "after_call", "after_write")

    def __init__(self,
                 name: str,
                 type: Type,
                 expire_after_call: timedelta = None,
                 expire_after_write: timedelta = None):
        """
        ValueSpec a specification of a persisted value.

        :param name: a user defined state name, used to represent this value. A name must be lower case, alphanumeric string
        without spaces, that starts with either a letter or an underscore (_)
        :param type: the value's Type. (see
        statefun.Type) :param expire_after_call: expire (remove) this value if a call to this function hasn't been
        made for the given duration.
        :param expire_after_write: expire this value if it wasn't written to for the given duration.
        """
        if not name:
            raise ValueError("name can not be missing.")
        if not name.isidentifier():
            raise ValueError(
                f"invalid name {name}. A spec name can only contains alphanumeric letters (a-z) and (0-9), "
                f"or underscores ( "
                f"_). A valid identifier cannot start with a number, or contain any spaces.")
        if iskeyword(name):
            forbidden = '\n'.join(kwlist)
            raise ValueError(
                f"invalid spec name {name} (Python SDK specifically). since {name} will result as an attribute on "
                f"context.store.\n"
                f"The following names are forbidden:\n {forbidden}")
        if not name.islower():
            raise ValueError(f"Only lower case names are allowed, {name} is given.")
        self.name = name
        if not type:
            raise ValueError("type can not be missing.")
        if not isinstance(type, Type):
            raise TypeError("type is not a StateFun type.")
        self.type = type
        if expire_after_call and expire_after_write:
            # both can not be set.
            raise ValueError("Either expire_after_call or expire_after_write can be set, but not both.")
        if expire_after_call:
            self.duration = int(expire_after_call.total_seconds() * 1000.0)
            self.after_call = True
            self.after_write = False
        elif expire_after_write:
            self.duration = int(expire_after_write.total_seconds() * 1000.0)
            self.after_call = False
            self.after_write = True
        else:
            self.duration = 0
            self.after_call = False
            self.after_write = False


def parse_typename(typename):
    """
    Parse a TypeName string into a namespace, type pair.
    :param typename: a string of the form <namespace>/<type>
    :return: a tuple of a namespace type.
    """
    if typename is None:
        raise ValueError("function type must be provided")
    idx = typename.rfind("/")
    if idx < 0:
        raise ValueError("function type must be of the from namespace/name")
    namespace = typename[:idx]
    if not namespace:
        raise ValueError("function type's namespace must not be empty")
    type = typename[idx + 1:]
    if not type:
        raise ValueError("function type's name must not be empty")
    return namespace, type


def simple_type(typename=None, serialize_fn=None, deserialize_fn=None) -> Type:
    """
    Create a user defined Type, simply by providing two functions. One for serializing a value to bytes, and one for
    deserializing the value from bytes.
    For example:

    import json
    tpe = simple_type("org.foo.bar/User", serialize_fn=json.dumps, deserialize_fn=json.loads)

    this defines a StateFun type (Type) of the typename org.foo.bar/User and it is a JSON object.

    :param typename: this type's TypeName.
    :param serialize_fn: a function that is able to serialize values of this type.
    :param deserialize_fn: a function that is able to deserialize bytes into values of this type.
    :return: a Type definition.
    """
    return SimpleType(typename, serialize_fn, deserialize_fn)


# ----------------------------------------------------------------------------------------------------------
# Internal
# ----------------------------------------------------------------------------------------------------------


class SimpleType(Type):
    __slots__ = ("typename", "_ser")

    def __init__(self, typename, serialize_fn, deserialize_fn):
        super().__init__(typename)
        if not serialize_fn:
            raise ValueError("serialize_fn is missing")
        if not deserialize_fn:
            raise ValueError("deserialize_fn is missing")
        self._ser = UserTypeSerializer(serialize_fn, deserialize_fn)

    def serializer(self) -> TypeSerializer:
        return self._ser


class UserTypeSerializer(TypeSerializer):
    __slots__ = ("serialize_fn", "deserialize_fn")

    def __init__(self, serialize_fn, deserialize_fn):
        self.serialize_fn = serialize_fn
        self.deserialize_fn = deserialize_fn

    def serialize(self, value):
        return self.serialize_fn(value)

    def deserialize(self, byte_string):
        return self.deserialize_fn(byte_string)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: kinesis-egress.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'kinesis-egress.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14kinesis-egress.proto\x12\x16io.statefun.sdk.egress\"l\n\x13KinesisEgressRecord\x12\x15\n\rpartition_key\x18\x01 \x01(\t\x12\x13\n\x0bvalue_bytes\x18\x02 \x01(\x0c\x12\x0e\n\x06stream\x18\x03 \x01(\t\x12\x19\n\x11\x65xplicit_hash_key\x18\x04 \x01(\tB>\n.org.apache.flink.statefun.sdk.egress.generatedP\x01Z\n.;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'kinesis_egress_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n.org.apache.flink.statefun.sdk.egress.generatedP\001Z\n.;protocol'
  _globals['_KINESISEGRESSRECORD']._serialized_start=48
  _globals['_KINESISEGRESSRECORD']._serialized_end=156
# @@protoc_insertion_point(module_scope)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

from statefun.core import Type, TypeSerializer
from statefun.types_pb2 import *


class ProtobufWrappingTypeSerializer(TypeSerializer):
    __slots__ = ("wrapper",)

    def __init__(self, wrapper):
        self.wrapper = wrapper

    def serialize(self, value):
        instance = self.wrapper()
        instance.value = value
        return instance.SerializeToString()

    def deserialize(self, string):
        instance = self.wrapper()
        instance.ParseFromString(string)
        return instance.value


class ProtobufWrappingType(Type):
    __slots__ = ("wrapper",)

    def __init__(self, typename, wrapper_message_type):
        super().__init__(typename)
        self.wrapper = wrapper_message_type

    def serializer(self) -> TypeSerializer:
        return ProtobufWrappingTypeSerializer(self.wrapper)


BoolType = ProtobufWrappingType("io.statefun.types/bool", BooleanWrapper)
IntType = ProtobufWrappingType("io.statefun.types/int", IntWrapper)
FloatType = ProtobufWrappingType("io.statefun.types/float", FloatWrapper)
LongType = ProtobufWrappingType("io.statefun.types/long", LongWrapper)
DoubleType = ProtobufWrappingType("io.statefun.types/double", DoubleWrapper)
StringType = ProtobufWrappingType("io.statefun.types/string", StringWrapper)

PY_TYPE_TO_WRAPPER_TYPE = {
    int: IntType,
    bool: BoolType,
    float: FloatType,
    str: StringType
}

WRAPPER_TYPE_TO_PY_TYPE = {
    IntType: int,
    BoolType: bool,
    FloatType: float,
    StringType: str
}



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import struct
import typing

from statefun.core import Type
from statefun.messages import EgressMessage

from statefun.kafka_egress_pb2 import KafkaProducerRecord
from statefun.kinesis_egress_pb2 import KinesisEgressRecord
from statefun.request_reply_pb2 import TypedValue


def kafka_egress_message(typename: str,
                         topic: str,
                         value: typing.Union[str, bytes, bytearray, int, float],
                         value_type: Type = None,
                         key: str = None):
    """
    Build a message that can be emitted to a Kafka generic egress.

    If a value_type is provided, then @value will be serialized according to the
    provided value_type's serializer. Otherwise we will try to convert @value to bytes
    if it is one of:
    - utf-8 string
    - bytes
    - bytearray
    - an int (as defined by Kafka's serialization format)
    - float (as defined by Kafka's serialization format)

    :param typename: the target egress to emit to (as defined in the module.yaml)
    :param topic: The Kafka destination topic for that record
    :param key: the utf8 encoded string key to produce (can be empty)
    :param value: the value to produce
    :param value_type: an optional hint to this value type.
    :return: A Protobuf message representing the record to be produced via the Kafka generic egress.
    """
    if not topic:
        raise ValueError("A destination Kafka topic is missing")
    if value is None:
        raise ValueError("Missing value")
    record = KafkaProducerRecord()
    record.topic = topic
    if value_type:
        ser = value_type.serializer()
        record.value_bytes = ser.serialize(value)
    elif isinstance(value, str):
        record.value_bytes = bytes(value, 'utf-8')
    elif isinstance(value, (bytes, bytearray)):
        record.value_bytes = bytes(value)
    elif isinstance(value, int):
        # see:
        # IntegerSerializer Javadoc
        # https://docs.confluent.io/current/clients/javadocs/org/apache/kafka/common/serialization/IntegerSerializer.html
        record.value_bytes = struct.pack('>i', value)
    elif isinstance(value, float):
        # see:
        # DoubleDeserializer Javadoc
        # https://docs.confluent.io/current/clients/javadocs/org/apache/kafka/common/serialization/DoubleDeserializer.html
        record.value_bytes = struct.pack('>d', value)
    else:
        raise TypeError("Unable to convert value to bytes.")
    if key is not None:
        record.key = key

    typed_value = TypedValue()
    typed_value.typename = "type.googleapis.com/io.statefun.sdk.egress.KafkaProducerRecord"
    typed_value.has_value = True
    typed_value.value = record.SerializeToString()

    return EgressMessage(typename, typed_value)


def kinesis_egress_message(typename: str,
                           stream: str,
                           value: typing.Union[str, bytes, bytearray],
                           partition_key: str,
                           value_type: typing.Union[None, Type] = None,
                           explicit_hash_key: str = None):
    """
    Build a message that can be emitted to a Kinesis generic egress.

    :param typename: the typename as specified in module.yaml
    :param stream: The AWS Kinesis destination stream for that record
    :param partition_key: the utf8 encoded string partition key to use
    :param value: the value to produce
    :param explicit_hash_key: a utf8 encoded string explicit hash key to use (can be empty)
    :param value_type: an optional hint to this value type
    :return: A Protobuf message representing the record to be produced to AWS Kinesis via the Kinesis generic egress.
    """
    if not stream:
        raise ValueError("Missing destination Kinesis stream")
    if value is None:
        raise ValueError("Missing value")
    if partition_key is None:
        raise ValueError("Missing partition key")
    record = KinesisEgressRecord()
    record.stream = stream
    if value_type:
        ser = value_type.serializer()
        record.value_bytes = ser.serialize(value)
    elif isinstance(value, str):
        record.value_bytes = bytes(value, 'utf-8')
    elif isinstance(value, (bytes, bytearray)):
        record.value_bytes = bytes(value)
    else:
        raise TypeError("Unable to convert value to bytes.")
    record.partition_key = partition_key
    if explicit_hash_key is not None:
        record.explicit_hash_key = explicit_hash_key

    typed_value = TypedValue()
    typed_value.typename = "type.googleapis.com/io.statefun.sdk.egress.KinesisEgressRecord"
    typed_value.has_value = True
    typed_value.value = record.SerializeToString()

    return EgressMessage(typename, typed_value)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import typing

from statefun.core import ValueSpec
from statefun.context import Context
from statefun.messages import Message
from statefun.storage import make_address_storage_spec, StorageSpec
import inspect


class StatefulFunction(object):
    __slots__ = ("fun", "storage_spec", "is_async")

    def __init__(self,
                 fun: typing.Callable[[Context, Message], None],
                 specs: StorageSpec,
                 is_async: bool):
        if fun is None:
            raise ValueError("function code is missing.")
        self.fun = fun
        if specs is None:
            raise ValueError("storage spec is missing.")
        self.storage_spec = specs
        self.is_async = is_async


class StatefulFunctions(object):
    __slots__ = ("_functions",)

    def __init__(self):
        self._functions = {}

    def register(self, typename: str, fun, specs: typing.Optional[typing.List[ValueSpec]] = None):
        """registers a StatefulFunction function instance, under the given namespace with the given function type. """
        if fun is None:
            raise ValueError("function instance must be provided")
        if not typename:
            raise ValueError("function typename must be provided")
        storage_spec = make_address_storage_spec(specs if specs else [])
        is_async = inspect.iscoroutinefunction(fun)
        sig = inspect.getfullargspec(fun)
        if len(sig.args) != 2:
            raise ValueError(
                f"The registered function {typename} does not expect a context and a message but rather {sig.args}.")
        self._functions[typename] = StatefulFunction(fun=fun, specs=storage_spec, is_async=is_async)

    def bind(self, typename, specs: typing.List[ValueSpec] = None):
        """wraps a StatefulFunction instance with a given namespace and type.
           for example:
            s = StatefulFunctions()

            @s.define("com.foo.bar/greeter")
            def greeter(context, message):
                print("Hi there")

            This would add an invokable stateful function that can accept messages
            sent to "com.foo.bar/greeter".
         """

        def wrapper(function):
            self.register(typename, function, specs)
            return function

        return wrapper

    def for_typename(self, typename: str) -> StatefulFunction:
        return self._functions[typename]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: types.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'types.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0btypes.proto\x12\x15io.statefun.sdk.types\"\x1f\n\x0e\x42ooleanWrapper\x12\r\n\x05value\x18\x01 \x01(\x08\"\x1b\n\nIntWrapper\x12\r\n\x05value\x18\x01 \x01(\x0f\"\x1d\n\x0c\x46loatWrapper\x12\r\n\x05value\x18\x01 \x01(\x02\"\x1c\n\x0bLongWrapper\x12\r\n\x05value\x18\x01 \x01(\x10\"\x1e\n\rDoubleWrapper\x12\r\n\x05value\x18\x01 \x01(\x01\"\x1e\n\rStringWrapper\x12\r\n\x05value\x18\x01 \x01(\tB=\n-org.apache.flink.statefun.sdk.types.generatedP\x01Z\n.;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'types_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n-org.apache.flink.statefun.sdk.types.generatedP\001Z\n.;protocol'
  _globals['_BOOLEANWRAPPER']._serialized_start=38
  _globals['_BOOLEANWRAPPER']._serialized_end=69
  _globals['_INTWRAPPER']._serialized_start=71
  _globals['_INTWRAPPER']._serialized_end=98
  _globals['_FLOATWRAPPER']._serialized_start=100
  _globals['_FLOATWRAPPER']._serialized_end=129
  _globals['_LONGWRAPPER']._serialized_start=131
  _globals['_LONGWRAPPER']._serialized_end=159
  _globals['_DOUBLEWRAPPER']._serialized_start=161
  _globals['_DOUBLEWRAPPER']._serialized_end=191
  _globals['_STRINGWRAPPER']._serialized_start=193
  _globals['_STRINGWRAPPER']._serialized_end=223
# @@protoc_insertion_point(module_scope)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import asyncio
import typing
from datetime import timedelta

import statefun.context
from statefun.core import parse_typename, ValueSpec, SdkAddress
from statefun.messages import Message, EgressMessage
from statefun.statefun_builder import StatefulFunctions, StatefulFunction

# generated function protocol
from statefun.request_reply_pb2 import ToFunction, FromFunction, Address, TypedValue
from statefun.storage import resolve, Cell

from dataclasses import dataclass


@dataclass
class DelayedMessage:
    is_cancellation: bool = None
    duration: int = None
    message: Message = None,
    cancellation_token: str = None


class UserFacingContext(statefun.context.Context):
    __slots__ = (
        "_self_address", "_outgoing_messages", "_outgoing_delayed_messages", "_outgoing_egress_messages", "_storage",
        "_caller")

    def __init__(self, address, storage):
        self._self_address = address
        self._outgoing_messages = []
        self._outgoing_delayed_messages: typing.List[DelayedMessage] = []
        self._outgoing_egress_messages = []
        self._storage = storage
        self._caller = None

    @property
    def address(self) -> SdkAddress:
        """

        :return: the address of the currently executing function. the address is of the form (typename, id)
        """
        return self._self_address

    @property
    def storage(self):
        return self._storage

    @property
    def caller(self):
        return self._caller

    def send(self, message: Message):
        """
        Send a message to a function.

        :param message: a message to send.
        """
        self._outgoing_messages.append(message)

    def send_after(self, duration: timedelta, message: Message, cancellation_token: str = ""):
        """
        Send a message to a target function after a specified delay.

        :param duration: the amount of time to wait before sending this message out.
        :param message: the message to send.
        :param cancellation_token: an optional cancellation token to associate with this message.
        """
        ms = int(duration.total_seconds() * 1000.0)
        record = DelayedMessage(is_cancellation=False, duration=ms, message=message,
                                cancellation_token=cancellation_token)
        self._outgoing_delayed_messages.append(record)

    def cancel_delayed_message(self, cancellation_token: str):
        """
        Cancel a delayed message (message that was sent using send_after) with a given token.

        Please note that this is a best-effort operation, since the message might have been already delivered.
        If the message was delivered, this is a no-op operation.
        """
        record = DelayedMessage(is_cancellation=True, cancellation_token=cancellation_token)
        self._outgoing_delayed_messages.append(record)

    def send_egress(self, message: EgressMessage):
        """
        Send a message to an egress.

        :param message: the EgressMessage to send.
        """
        self._outgoing_egress_messages.append(message)


# -------------------------------------------------------------------------------------------------------------------
# Protobuf Helpers
# -------------------------------------------------------------------------------------------------------------------

def sdk_address_from_pb(addr: Address) -> typing.Optional[SdkAddress]:
    if not addr or (not addr.namespace and not addr.type and not addr.id):
        return None
    return SdkAddress(namespace=addr.namespace,
                      name=addr.type,
                      id=addr.id,
                      typename=f"{addr.namespace}/{addr.type}")


# noinspection PyProtectedMember
def collect_success(ctx: UserFacingContext) -> FromFunction:
    pb_from_function = FromFunction()
    pb_invocation_result = pb_from_function.invocation_result
    collect_messages(ctx._outgoing_messages, pb_invocation_result)
    collect_delayed(ctx._outgoing_delayed_messages, pb_invocation_result)
    collect_egress(ctx._outgoing_egress_messages, pb_invocation_result)
    collect_mutations(ctx._storage._cells, pb_invocation_result)
    return pb_from_function


def collect_failure(missing_state_specs: typing.List[ValueSpec]) -> FromFunction:
    pb_from_function = FromFunction()
    incomplete_context = pb_from_function.incomplete_invocation_context
    missing_values = incomplete_context.missing_values
    for state_spec in missing_state_specs:
        missing_value = missing_values.add()
        missing_value.state_name = state_spec.name
        missing_value.type_typename = state_spec.type.typename

        protocol_expiration_spec = FromFunction.ExpirationSpec()
        if not state_spec.after_write and not state_spec.after_call:
            protocol_expiration_spec.mode = FromFunction.ExpirationSpec.ExpireMode.NONE
        else:
            protocol_expiration_spec.expire_after_millis = state_spec.duration
            if state_spec.after_call:
                protocol_expiration_spec.mode = FromFunction.ExpirationSpec.ExpireMode.AFTER_INVOKE
            elif state_spec.after_write:
                protocol_expiration_spec.mode = FromFunction.ExpirationSpec.ExpireMode.AFTER_WRITE
            else:
                raise ValueError("Unexpected state expiration mode.")
        missing_value.expiration_spec.CopyFrom(protocol_expiration_spec)
    return pb_from_function


def collect_messages(messages: typing.List[Message], pb_invocation_result):
    pb_outgoing_messages = pb_invocation_result.outgoing_messages
    for message in messages:
        outgoing = pb_outgoing_messages.add()

        namespace, type = parse_typename(message.target_typename)
        outgoing.target.namespace = namespace
        outgoing.target.type = type
        outgoing.target.id = message.target_id
        outgoing.argument.CopyFrom(message.typed_value)


def collect_delayed(delayed_messages: typing.List[DelayedMessage], invocation_result):
    delayed_invocations = invocation_result.delayed_invocations
    for delayed_message in delayed_messages:
        outgoing = delayed_invocations.add()

        if delayed_message.is_cancellation:
            # handle cancellation
            outgoing.cancellation_token = delayed_message.cancellation_token
            outgoing.is_cancellation_request = True
        else:
            message = delayed_message.message
            namespace, type = parse_typename(message.target_typename)

            outgoing.target.namespace = namespace
            outgoing.target.type = type
            outgoing.target.id = message.target_id
            outgoing.delay_in_ms = delayed_message.duration
            outgoing.argument.CopyFrom(message.typed_value)
            if delayed_message.cancellation_token is not None:
                outgoing.cancellation_token = delayed_message.cancellation_token


def collect_cancellations(tokens: typing.List[str], invocation_result):
    outgoing_cancellations = invocation_result.outgoing_delay_cancellations
    for token in tokens:
        if token:
            delay_cancelltion = outgoing_cancellations.add()
            delay_cancelltion.cancellation_token = token


def collect_egress(egresses: typing.List[EgressMessage], invocation_result):
    outgoing_egresses = invocation_result.outgoing_egresses
    for message in egresses:
        outgoing = outgoing_egresses.add()

        namespace, type = parse_typename(message.typename)
        outgoing.egress_namespace = namespace
        outgoing.egress_type = type
        outgoing.argument.CopyFrom(message.typed_value)


def collect_mutations(cells: typing.Dict[str, Cell], invocation_result):
    for key, cell in cells.items():
        if not cell.dirty:
            continue
        mutation = invocation_result.state_mutations.add()
        mutation.state_name = key
        val: typing.Optional[TypedValue] = cell.typed_value
        if val is None:
            # it is deleted.
            mutation.mutation_type = FromFunction.PersistedValueMutation.MutationType.Value('DELETE')
        else:
            mutation.mutation_type = FromFunction.PersistedValueMutation.MutationType.Value('MODIFY')
            mutation.state_value.CopyFrom(val)


# --------------------------------------------------------------------------------------------------------------------
# The main Request Reply Handler.
# --------------------------------------------------------------------------------------------------------------------

class RequestReplyHandler(object):
    def __init__(self, functions: StatefulFunctions):
        if not functions:
            raise ValueError("functions must be provided.")
        self.functions = functions

    def handle_sync(self, request_bytes: typing.Union[str, bytes, bytearray]) -> bytes:
        return asyncio.run(self.handle_async(request_bytes))

    async def handle_async(self, request_bytes: typing.Union[str, bytes, bytearray]) -> bytes:
        # parse
        pb_to_function = ToFunction()
        pb_to_function.ParseFromString(request_bytes)
        # target address
        pb_target_address = pb_to_function.invocation.target
        sdk_address = sdk_address_from_pb(pb_target_address)
        # target stateful function
        target_fn: StatefulFunction = self.functions.for_typename(sdk_address.typename)
        if not target_fn:
            raise ValueError(f"Unable to find a function of type {sdk_address.typename}")
        # resolve state
        res = resolve(target_fn.storage_spec, sdk_address.typename, pb_to_function.invocation.state)
        if res.missing_specs:
            pb_from_function = collect_failure(res.missing_specs)
            return pb_from_function.SerializeToString()
        # invoke the batch
        ctx = UserFacingContext(sdk_address, res.storage)
        fun = target_fn.fun
        pb_batch = pb_to_function.invocation.invocations
        if target_fn.is_async:
            for pb_invocation in pb_batch:
                msg = Message(target_typename=sdk_address.typename, target_id=sdk_address.id,
                              typed_value=pb_invocation.argument)
                ctx._caller = sdk_address_from_pb(pb_invocation.caller)
                # await for an async function to complete.
                # noinspection PyUnresolvedReferences
                await fun(ctx, msg)
        else:
            for pb_invocation in pb_batch:
                msg = Message(target_typename=sdk_address.typename, target_id=sdk_address.id,
                              typed_value=pb_invocation.argument)
                ctx._caller = sdk_address_from_pb(pb_invocation.caller)
                # we need to call the function directly ¯\_(ツ)_/¯
                fun(ctx, msg)
        # collect the results
        pb_from_function = collect_success(ctx)
        return pb_from_function.SerializeToString()



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

# type API
from statefun.core import TypeSerializer, Type, simple_type
from statefun.core import ValueSpec
from statefun.core import SdkAddress

# wrapper types
from statefun.wrapper_types import BoolType, IntType, FloatType, DoubleType, LongType, StringType

# messaging
from statefun.messages import Message, EgressMessage, message_builder, egress_message_builder

# egress io
from statefun.egress_io import kafka_egress_message, kinesis_egress_message

# context
from statefun.context import Context

# statefun builder
from statefun.statefun_builder import StatefulFunctions

# request reply protocol handler
from statefun.request_reply_v3 import RequestReplyHandler

# utilits
from statefun.utils import make_protobuf_type, make_json_type



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import abc
import typing
from datetime import timedelta

from statefun.core import SdkAddress
from statefun.messages import Message, EgressMessage


class Context(abc.ABC):
    __slots__ = ()

    @property
    @abc.abstractmethod
    def address(self) -> SdkAddress:
        """

        :return: the address of the currently executing function. the address is of the form (typename, id)
        """
        pass

    @property
    @abc.abstractmethod
    def storage(self):
        """

        :return: the address scoped storage.
        """
        pass

    @property
    @abc.abstractmethod
    def caller(self) -> typing.Union[None, SdkAddress]:
        """

        :return: the address of the caller or None if this function was triggered by the ingress.
        """
        pass

    def send(self, message: Message):
        """
        Send a message to a function.

        :param message: a message to send.
        """
        pass

    def send_after(self, duration: timedelta, message: Message, cancellation_token: str = ""):
        """
        Send a message to a target function after a specified delay.

        :param duration: the amount of time to wait before sending this message out.
        :param message: the message to send.
        :param cancellation_token: an optional cancellation token to associate with this message.
        """
        pass

    def cancel_delayed_message(self, cancellation_token: str):
        """
        Cancel a delayed message (message that was sent using send_after) with a given token.

        Please note that this is a best-effort operation, since the message might have been already delivered.
        If the message was delivered, this is a no-op operation.
        """
        pass

    def send_egress(self, message: EgressMessage):
        """
        Send a message to an egress.

        :param message: the EgressMessage to send.
        """
        pass



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import typing
from dataclasses import dataclass

from statefun.core import ValueSpec, Type
from statefun.request_reply_pb2 import TypedValue, ToFunction

from statefun.utils import to_typed_value, from_typed_value


@dataclass(frozen=True, repr=True, eq=True, order=False)
class StorageSpec:
    make_instance: typing.Any
    names: typing.FrozenSet[str]
    specs: typing.List[ValueSpec]


@dataclass(frozen=True, eq=False, repr=False, order=False)
class Resolution:
    missing_specs: typing.Union[None, typing.List[ValueSpec]]
    storage: typing.Union[None, typing.Any]  # this holds the storage instance.


class Cell(object):
    __slots__ = ("tpe", "typed_value", "dirty")

    def __init__(self, tpe: Type, typed_value: typing.Optional[TypedValue]):
        # read only
        self.tpe = tpe
        # mutable
        self.typed_value = typed_value
        self.dirty = False

    def get(self):
        typed_value = self.typed_value
        return from_typed_value(self.tpe, typed_value)

    def set(self, val):
        if val is None:
            raise ValueError('provided value must not be None. To delete a value, please use del.')
        tpe = self.tpe
        typed_value = to_typed_value(tpe, val)
        self.typed_value = typed_value
        self.dirty = True

    def delete(self):
        self.typed_value = None
        self.dirty = True


# self.cells: typing.Dict[str, Cell] = {name: Cell(name, tpe, vals[name]) for name, tpe in types.items()}


def storage_constructor(self, cells: typing.Dict[str, Cell], typename: str):
    self._cells = cells
    self._typename = typename


def storage_missing_attribute(self, attr):
    raise AttributeError("'{}' is not a registered ValueSpec for the function '{}'".format(attr, self._typename))


def property_named(name):
    """
    Creates a property, that delegates all the operations to an instance
    of a HiddenClass, that is expected to be member of target class (the class that
    this property will be part of).

    :param name: the name of this field
    :return: a property as described above.
    """

    def fget(self):
        cell: Cell = self._cells.get(name)
        if not cell:
            raise AttributeError(name)
        return cell.get()

    def fset(self, val):
        cell: Cell = self._cells.get(name)
        if not cell:
            raise AttributeError(name)
        cell.set(val)

    def fdel(self):
        cell: Cell = self._cells.get(name)
        if not cell:
            raise AttributeError(name)
        cell.delete()

    return property(fget, fset, fdel)


def make_address_storage_spec(specs: typing.List[ValueSpec]) -> StorageSpec:
    """
    Creates an StorageSpec from user supplied value specs.
    :param specs: a list of specs as supplied by the user.
    :return: a StorageSpec.
    """
    props = {
        "__init__": storage_constructor,
        "__getattr__": storage_missing_attribute,
        "__slots__": ["_cells", "_typename"]}
    for spec in specs:
        if spec.name in props:
            raise ValueError("duplicate registered value name: " + spec.name)
        props[spec.name] = property_named(spec.name)
    cls = type("GeneratedAddressedScopedStorage", (object,), props)
    return StorageSpec(make_instance=cls,
                       names=frozenset(spec.name for spec in specs),
                       specs=specs)


def resolve(storage: StorageSpec,
            typename: str,
            values: typing.List[ToFunction.PersistedValue]) -> Resolution:
    """
    Resolve the registered specs and the actually received values.

    :param storage: a storage factory
    :param typename: the typename of the function under invocation
    :param values: the actually received values
    :return: a Resolution result, that might have either a list of missing specs
    (specs that were defined by the user but didn't arrived from StateFun) or a
     successful resolution, with an instance of an addressed scoped storage.
    """
    #
    # index the actually received values (TypedValue) by name.
    #
    received: typing.Dict[str, TypedValue] = {state.state_name: state.state_value for state in values}
    #
    # see if any of the specs are missing.
    #
    missing_keys = storage.names - received.keys()
    if missing_keys:
        # keep the missing specs in exactly the same order as they were originally defined.
        # This is not strictly required from the protocol point of view, but it makes it a bit easier
        # to troubleshoot.
        missing = [spec for spec in storage.specs if spec.name in missing_keys]
        return Resolution(missing_specs=missing, storage=None)
    else:
        cells: typing.Dict[str, Cell] = {spec.name: Cell(tpe=spec.type, typed_value=received[spec.name]) for spec in
                                         storage.specs}
        s = storage.make_instance(cells, typename)
        return Resolution(missing_specs=None, storage=s)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: request-reply.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'request-reply.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x13request-reply.proto\x12\x18io.statefun.sdk.reqreply\"6\n\x07\x41\x64\x64ress\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\n\n\x02id\x18\x03 \x01(\t\"@\n\nTypedValue\x12\x10\n\x08typename\x18\x01 \x01(\t\x12\x11\n\thas_value\x18\x02 \x01(\x08\x12\r\n\x05value\x18\x03 \x01(\x0c\"\x9c\x04\n\nToFunction\x12Q\n\ninvocation\x18\x64 \x01(\x0b\x32;.io.statefun.sdk.reqreply.ToFunction.InvocationBatchRequestH\x00\x1a_\n\x0ePersistedValue\x12\x12\n\nstate_name\x18\x01 \x01(\t\x12\x39\n\x0bstate_value\x18\x02 \x01(\x0b\x32$.io.statefun.sdk.reqreply.TypedValue\x1aw\n\nInvocation\x12\x31\n\x06\x63\x61ller\x18\x01 \x01(\x0b\x32!.io.statefun.sdk.reqreply.Address\x12\x36\n\x08\x61rgument\x18\x02 \x01(\x0b\x32$.io.statefun.sdk.reqreply.TypedValue\x1a\xd5\x01\n\x16InvocationBatchRequest\x12\x31\n\x06target\x18\x01 \x01(\x0b\x32!.io.statefun.sdk.reqreply.Address\x12\x42\n\x05state\x18\x02 \x03(\x0b\x32\x33.io.statefun.sdk.reqreply.ToFunction.PersistedValue\x12\x44\n\x0binvocations\x18\x03 \x03(\x0b\x32/.io.statefun.sdk.reqreply.ToFunction.InvocationB\t\n\x07request\"\xbc\r\n\x0c\x46romFunction\x12V\n\x11invocation_result\x18\x64 \x01(\x0b\x32\x39.io.statefun.sdk.reqreply.FromFunction.InvocationResponseH\x00\x12k\n\x1dincomplete_invocation_context\x18\x65 \x01(\x0b\x32\x42.io.statefun.sdk.reqreply.FromFunction.IncompleteInvocationContextH\x00\x1a\xf2\x01\n\x16PersistedValueMutation\x12\x61\n\rmutation_type\x18\x01 \x01(\x0e\x32J.io.statefun.sdk.reqreply.FromFunction.PersistedValueMutation.MutationType\x12\x12\n\nstate_name\x18\x02 \x01(\t\x12\x39\n\x0bstate_value\x18\x03 \x01(\x0b\x32$.io.statefun.sdk.reqreply.TypedValue\"&\n\x0cMutationType\x12\n\n\x06\x44\x45LETE\x10\x00\x12\n\n\x06MODIFY\x10\x01\x1aw\n\nInvocation\x12\x31\n\x06target\x18\x01 \x01(\x0b\x32!.io.statefun.sdk.reqreply.Address\x12\x36\n\x08\x61rgument\x18\x02 \x01(\x0b\x32$.io.statefun.sdk.reqreply.TypedValue\x1a\xd0\x01\n\x11\x44\x65layedInvocation\x12\x1f\n\x17is_cancellation_request\x18\n \x01(\x08\x12\x1a\n\x12\x63\x61ncellation_token\x18\x0b \x01(\t\x12\x13\n\x0b\x64\x65lay_in_ms\x18\x01 \x01(\x03\x12\x31\n\x06target\x18\x02 \x01(\x0b\x32!.io.statefun.sdk.reqreply.Address\x12\x36\n\x08\x61rgument\x18\x03 \x01(\x0b\x32$.io.statefun.sdk.reqreply.TypedValue\x1av\n\rEgressMessage\x12\x18\n\x10\x65gress_namespace\x18\x01 \x01(\t\x12\x13\n\x0b\x65gress_type\x18\x02 \x01(\t\x12\x36\n\x08\x61rgument\x18\x03 \x01(\x0b\x32$.io.statefun.sdk.reqreply.TypedValue\x1a\xe2\x02\n\x12InvocationResponse\x12V\n\x0fstate_mutations\x18\x01 \x03(\x0b\x32=.io.statefun.sdk.reqreply.FromFunction.PersistedValueMutation\x12L\n\x11outgoing_messages\x18\x02 \x03(\x0b\x32\x31.io.statefun.sdk.reqreply.FromFunction.Invocation\x12U\n\x13\x64\x65layed_invocations\x18\x03 \x03(\x0b\x32\x38.io.statefun.sdk.reqreply.FromFunction.DelayedInvocation\x12O\n\x11outgoing_egresses\x18\x04 \x03(\x0b\x32\x34.io.statefun.sdk.reqreply.FromFunction.EgressMessage\x1a\xb8\x01\n\x0e\x45xpirationSpec\x12N\n\x04mode\x18\x01 \x01(\x0e\x32@.io.statefun.sdk.reqreply.FromFunction.ExpirationSpec.ExpireMode\x12\x1b\n\x13\x65xpire_after_millis\x18\x02 \x01(\x03\"9\n\nExpireMode\x12\x08\n\x04NONE\x10\x00\x12\x0f\n\x0b\x41\x46TER_WRITE\x10\x01\x12\x10\n\x0c\x41\x46TER_INVOKE\x10\x02\x1a\x8f\x01\n\x12PersistedValueSpec\x12\x12\n\nstate_name\x18\x01 \x01(\t\x12N\n\x0f\x65xpiration_spec\x18\x02 \x01(\x0b\x32\x35.io.statefun.sdk.reqreply.FromFunction.ExpirationSpec\x12\x15\n\rtype_typename\x18\x03 \x01(\t\x1ap\n\x1bIncompleteInvocationContext\x12Q\n\x0emissing_values\x18\x01 \x03(\x0b\x32\x39.io.statefun.sdk.reqreply.FromFunction.PersistedValueSpecB\n\n\x08responseB@\n0org.apache.flink.statefun.sdk.reqreply.generatedP\x01Z\n.;protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'request_reply_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n0org.apache.flink.statefun.sdk.reqreply.generatedP\001Z\n.;protocol'
  _globals['_ADDRESS']._serialized_start=49
  _globals['_ADDRESS']._serialized_end=103
  _globals['_TYPEDVALUE']._serialized_start=105
  _globals['_TYPEDVALUE']._serialized_end=169
  _globals['_TOFUNCTION']._serialized_start=172
  _globals['_TOFUNCTION']._serialized_end=712
  _globals['_TOFUNCTION_PERSISTEDVALUE']._serialized_start=269
  _globals['_TOFUNCTION_PERSISTEDVALUE']._serialized_end=364
  _globals['_TOFUNCTION_INVOCATION']._serialized_start=366
  _globals['_TOFUNCTION_INVOCATION']._serialized_end=485
  _globals['_TOFUNCTION_INVOCATIONBATCHREQUEST']._serialized_start=488
  _globals['_TOFUNCTION_INVOCATIONBATCHREQUEST']._serialized_end=701
  _globals['_FROMFUNCTION']._serialized_start=715
  _globals['_FROMFUNCTION']._serialized_end=2439
  _globals['_FROMFUNCTION_PERSISTEDVALUEMUTATION']._serialized_start=929
  _globals['_FROMFUNCTION_PERSISTEDVALUEMUTATION']._serialized_end=1171
  _globals['_FROMFUNCTION_PERSISTEDVALUEMUTATION_MUTATIONTYPE']._serialized_start=1133
  _globals['_FROMFUNCTION_PERSISTEDVALUEMUTATION_MUTATIONTYPE']._serialized_end=1171
  _globals['_FROMFUNCTION_INVOCATION']._serialized_start=1173
  _globals['_FROMFUNCTION_INVOCATION']._serialized_end=1292
  _globals['_FROMFUNCTION_DELAYEDINVOCATION']._serialized_start=1295
  _globals['_FROMFUNCTION_DELAYEDINVOCATION']._serialized_end=1503
  _globals['_FROMFUNCTION_EGRESSMESSAGE']._serialized_start=1505
  _globals['_FROMFUNCTION_EGRESSMESSAGE']._serialized_end=1623
  _globals['_FROMFUNCTION_INVOCATIONRESPONSE']._serialized_start=1626
  _globals['_FROMFUNCTION_INVOCATIONRESPONSE']._serialized_end=1980
  _globals['_FROMFUNCTION_EXPIRATIONSPEC']._serialized_start=1983
  _globals['_FROMFUNCTION_EXPIRATIONSPEC']._serialized_end=2167
  _globals['_FROMFUNCTION_EXPIRATIONSPEC_EXPIREMODE']._serialized_start=2110
  _globals['_FROMFUNCTION_EXPIRATIONSPEC_EXPIREMODE']._serialized_end=2167
  _globals['_FROMFUNCTION_PERSISTEDVALUESPEC']._serialized_start=2170
  _globals['_FROMFUNCTION_PERSISTEDVALUESPEC']._serialized_end=2313
  _globals['_FROMFUNCTION_INCOMPLETEINVOCATIONCONTEXT']._serialized_start=2315
  _globals['_FROMFUNCTION_INCOMPLETEINVOCATIONCONTEXT']._serialized_end=2427
# @@protoc_insertion_point(module_scope)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import typing

import statefun.wrapper_types as wrapper_types
from statefun.core import Type
from statefun.request_reply_pb2 import TypedValue
from statefun.utils import to_typed_value


class Message(object):
    __slots__ = ("target_typename", "target_id", "typed_value")

    def __init__(self, target_typename: str, target_id: str, typed_value: TypedValue):
        """
        A Stateful Functions Message.

        :param target_typename: The TypeName represented as a string of the form <namespace>/<name> of the
        target function.
        :param target_id: The id of the target function
        :param typed_value: The internal protobuf representation of the typed_value.
        """
        if not target_typename:
            raise ValueError("target_typename can not be missing")
        if not target_id:
            raise ValueError("target_id can not be missing")
        if not typed_value:
            raise ValueError("typed_value can not be missing")
        self.target_typename = target_typename
        self.target_id = target_id
        self.typed_value = typed_value

    def is_int(self):
        return self.is_type(wrapper_types.IntType)

    def as_int(self):
        return self.as_type(wrapper_types.IntType)

    def is_bool(self):
        return self.is_type(wrapper_types.BoolType)

    def as_bool(self) -> typing.Optional[bool]:
        return self.as_type(wrapper_types.BoolType)

    def is_long(self):
        return self.is_type(wrapper_types.LongType)

    def as_long(self) -> typing.Optional[int]:
        return self.as_type(wrapper_types.LongType)

    def is_string(self):
        return self.is_type(wrapper_types.StringType)

    def as_string(self) -> typing.Optional[str]:
        return self.as_type(wrapper_types.StringType)

    def is_float(self):
        return self.is_type(wrapper_types.FloatType)

    def as_float(self) -> typing.Optional[float]:
        return self.as_type(wrapper_types.FloatType)

    def is_double(self):
        return self.is_type(wrapper_types.DoubleType)

    def as_double(self) -> typing.Optional[float]:
        return self.as_type(wrapper_types.DoubleType)

    def is_type(self, tpe: Type) -> bool:
        return self.typed_value.typename == tpe.typename

    def value_typename(self) -> str:
        return self.typed_value.typename

    def raw_value(self) -> typing.Optional[bytes]:
        tv = self.typed_value
        return tv.value if tv.has_value else None

    def as_type(self, tpe: Type):
        tv = self.typed_value
        if tv.has_value:
            serializer = tpe.serializer()
            return serializer.deserialize(tv.value)
        else:
            return None


class EgressMessage(object):
    __slots__ = ("typename", "typed_value")

    def __init__(self, typename: str, typed_value: TypedValue):
        if not typename:
            raise ValueError("typename is missing")
        if not typed_value:
            raise ValueError("value is missing")
        self.typename = typename
        self.typed_value = typed_value

    def value_typename(self) -> str:
        return self.typed_value.typename

    def raw_value(self) -> typing.Optional[bytes]:
        tv = self.typed_value
        if tv.has_value:
            return tv.value
        else:
            return None


PRIMITIVE_GETTERS = {"int_value": wrapper_types.IntType,
                     "float_value": wrapper_types.FloatType,
                     "long_value": wrapper_types.LongType,
                     "str_value": wrapper_types.StringType,
                     "double_value": wrapper_types.DoubleType,
                     "bool_value": wrapper_types.BoolType}


def message_builder(target_typename: str, target_id: str, **kwargs) -> Message:
    """
    Build a Message that can be sent to any other function.
    :param target_typename: The TypeName represented as a string of the form <namespace>/<name> of the
           target function.
    :param target_id: The id of the target function
    :param kwargs: This specify the value type to attach to this message. The following arguments are supported:
                    int_value=<an int>,
                    float_value=<a float>
                    long_value=<a signed 64 bit integer>
                    str_value=<str>
                    double_value=<double>
                    bool_value=<bool>
                    ...
                    value=<arbitrary value>, value_type=<a StateFun Type for this value>
    :return: A Message object, that can be sent.
    """
    if len(kwargs) == 2:
        value, value_type = kwargs["value"], kwargs["value_type"]
    elif len(kwargs) == 1:
        # expecting: <type>_value : value
        # for example one of the following:
        #              int_value=1
        #              str_value="hello world"
        #              long_value= 5511
        type_keyword, value = next(iter(kwargs.items()))
        value_type = PRIMITIVE_GETTERS.get(type_keyword)
    else:
        raise TypeError(f"Wrong number of value keywords given: {kwargs}, there must be exactly one of:"
                        f"\nint_value=.."
                        f"\nfloat_value.."
                        f"\netc'"
                        f"\nor:"
                        f"\nvalue=.. ,value_type=.. ")
    if value is None:
        raise ValueError("value can not be missing")
    if not value_type:
        raise ValueError(
            "Could not deduce the value type, please specify the type explicitly. via passing: value=<the value>, "
            "value_type=<the type>")
    typed_value = to_typed_value(type=value_type, value=value)
    return Message(target_typename=target_typename, target_id=target_id, typed_value=typed_value)


def egress_message_builder(target_typename: str, value: typing.Any, value_type: Type):
    """
    Create a generic egress record.

    To use Kafka specific egress please use kafka_egress_message(), and for Kinesis please use
    kinesis_egress_message().
    """
    if not target_typename:
        raise ValueError("target typename is missing")
    if value is None:
        raise ValueError("value can not be missing")
    typed_value = to_typed_value(type=value_type, value=value)
    return EgressMessage(typename=target_typename, typed_value=typed_value)



################################################################################
#  Licensed to the Apache Software Foundation (ASF) under one
#  or more contributor license agreements.  See the NOTICE file
#  distributed with this work for additional information
#  regarding copyright ownership.  The ASF licenses this file
#  to you under the Apache License, Version 2.0 (the
#  "License"); you may not use this file except in compliance
#  with the License.  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
import typing
import json

from statefun.request_reply_pb2 import TypedValue
from statefun.core import Type, simple_type


def to_typed_value_deduce(value: typing.Union[str, bytes, bytearray],
                          value_type: typing.Union[None, Type] = None) -> bytes:
    if value_type:
        ser = value_type.serializer()
        return ser.serialize(value)
    if isinstance(value, (bytes, bytearray)):
        return bytes(value)
    try:
        return value.encode('utf-8')
    except AttributeError:
        pass
    raise TypeError(
        "Unable to convert value to bytes. strings, bytes, and anytype that has __bytes__ attribute is supported.")


def from_typed_value(tpe: Type, typed_value: typing.Optional[TypedValue]):
    if not typed_value or not typed_value.has_value:
        return None
    if tpe.typename != typed_value.typename:
        raise TypeError(
            f"Type mismatch: "
            f"type {typed_value.typename} (remote), yet it is specified locally "
            f"as {tpe.typename} ")
    ser = tpe.serializer()
    return ser.deserialize(typed_value.value)


def to_typed_value(type, value):
    typed_value = TypedValue()
    typed_value.typename = type.typename
    if value is None:
        typed_value.has_value = False
        return typed_value
    typed_value.has_value = True
    ser = type.serializer()
    typed_value.value = ser.serialize(value)
    return typed_value


def make_protobuf_type(cls, namespace: str = None) -> Type:
    """
    Create a StateFun type that is backed by Protobuf.
    """

    def deserialize_fn(b):
        v = cls()
        v.ParseFromString(b)
        return v

    def serialize_fn(v):
        return v.SerializeToString()

    if not namespace:
        namespace = "type.googleapis.com"
    if not cls:
        raise ValueError("The Protobuf generated class is missing.")
    name = cls().DESCRIPTOR.full_name
    if not name:
        raise TypeError("Unable to deduce the Protobuf Message full name.")
    return simple_type(typename=f"{namespace}/{name}", serialize_fn=serialize_fn, deserialize_fn=deserialize_fn)


def _serialize_json_utf8(obj) -> bytes:
    """
    serialize the given object as a JSON utf-8 bytes.
    """
    str = json.dumps(obj, ensure_ascii=False)
    return str.encode('utf-8')


def make_json_type(typename: str) -> Type:
    """
    Create a StateFun type named @typename, that is backed by Python's json module.
    """
    return simple_type(typename=typename,
                       serialize_fn=_serialize_json_utf8,
                       deserialize_fn=json.loads)
